from SpellChecker.TrieNode cimport TrieNode


cdef class Trie:

    cdef TrieNode __root_node

    cpdef insert(self, str word)
    cpdef bint search(self, str word)
    cpdef bint startsWith(self, str prefix)
    cpdef TrieNode getTrieNode(self, str word)



cdef class TrieNode:

    cdef dict __children
    cdef bint __isWord

    cpdef TrieNode getChild(self, str ch)
    cpdef addChild(self, str ch, TrieNode child)
    cpdef str childrenToString(self)
    cpdef bint isWord(self)
    cpdef setIsWord(self, bint isWord)



from collections import deque

import pkg_resources
from Corpus.Sentence cimport Sentence
from Dictionary.Word cimport Word
from MorphologicalAnalysis.FsmMorphologicalAnalyzer cimport FsmMorphologicalAnalyzer
from NGram.NGram cimport NGram

from SpellChecker.Candidate cimport Candidate
from SpellChecker.NGramSpellChecker cimport NGramSpellChecker
from SpellChecker.SpellCheckerParameter cimport SpellCheckerParameter
from SpellChecker.Trie cimport Trie
from SpellChecker.TrieCandidate cimport TrieCandidate
from SpellChecker.TrieNode cimport TrieNode

cdef class TrieBasedSpellChecker(NGramSpellChecker):

    def __init__(self,
                 fsm: FsmMorphologicalAnalyzer,
                 nGram: NGram,
                 parameter: SpellCheckerParameter):
        """
        A constructor of {@link TrieBasedSpellChecker} class which takes a {@link FsmMorphologicalAnalyzer}, an {@link NGram}
        and a {@link SpellCheckerParameter} as inputs. Then, calls its super class {@link NGramSpellChecker} with given inputs.
        :param fsm: {@link FsmMorphologicalAnalyzer} type input.
        :param nGram: {@link NGram} type input.
        :param parameter: {@link SpellCheckerParameter} type input.
        """
        super().__init__(fsm,
                         nGram,
                         parameter)
        self.loadTrieDictionaries()
        self.prepareTrie()

    cpdef loadTrieDictionaries(self):
        """
        This method also loads generated words from a file.
        """
        cdef list lines
        cdef str line
        self.__generated_words = []
        input_file = open(pkg_resources.resource_filename(__name__, 'data/generated_words.txt'), "r", encoding="utf8")
        lines = input_file.readlines()
        for line in lines:
            self.__generated_words.append(line.strip())
        input_file.close()

    cpdef prepareTrie(self):
        """
        Populates a Trie data structure with a set of strings generated by the morphological analyzer.
        """
        cdef str word
        self.__trie = Trie()
        for word in self.__generated_words:
            self.__trie.insert(word)

    cpdef list candidateList(self, Word word, Sentence sentence):
        """
        Checks if the trie object contains the candidate.
        If it does, it compares the penalty of the candidate in the results with the penalty of the candidate variable.
        If the current penalty of the candidate in the results is less than or equal to the current penalty
        of the candidate variable, it removes the first element of the candidates ArrayDeque.
        If not, it adds the candidate variable to the results ArrayList and removes the first element.
        If the trie object does not contain the candidate, it checks if the current penalty and the current
        index of the candidate variable and if they are not at the limit,
        it moves the index one character to the right and generates new candidates.
        :param word: the word to search for candidates for
        :param sentence: the sentence containing the word
        :return: an ArrayList of Candidate objects representing the potential candidates
        """
        cdef list results
        cdef float penalty_limit
        cdef TrieCandidate candidate, new_candidate
        cdef str candidate_name
        cdef int item_index
        candidates = deque()
        results = []
        candidates.append(TrieCandidate(word.getName(), -1, 0))
        penalty_limit = min(word.charCount() / 2.0, 3.0)
        while len(candidates) != 0:
            candidate = candidates[0]
            candidate_name = candidate.getName()
            if self.__trie.search(candidate_name):
                item_index = self.searchCandidates(results, candidate)
                if item_index != -1 and results[item_index].getCurrentPenalty() <= candidate.getCurrentPenalty():
                    candidates.popleft()
                else:
                    results.append(candidates.popleft())
            else:
                if candidate.getCurrentPenalty() > penalty_limit - 1 or candidate.getCurrentIndex() >= len(
                        candidate_name) - 1:
                    candidates.popleft()
                else:
                    candidate = candidates.popleft()
                    candidate.nextIndex()
                    candidates.append(
                        TrieCandidate(candidate.getName(), candidate.getCurrentIndex(), candidate.getCurrentPenalty()))
                    new_candidates = self.generateTrieCandidates(candidate)
                    for new_candidate in new_candidates:
                        candidates.append(new_candidate)
        return results

    cpdef int searchCandidates(self, list result, TrieCandidate candidate):
        cdef int i
        for i in range(len(result)):
            if result[i].getName() == candidate.getName():
                return i
        return -1

    cpdef list generateTrieCandidates(self, TrieCandidate candidate):
        """
        Generates a set of candidates based on a given TrieCandidate.
        The generated candidates are created by applying a set of operations to
        the input TrieCandidate. The possible operations are:
        De-asciification: replacing certain ASCII characters with their non-ASCII counterparts.
        Substitution: replacing a single character with another character.
        Insertion: adding a single character to the string.
        Deletion: removing a single character from the string.
        Transposition: swapping the positions of two adjacent characters in the string.
        :param candidate: the input TrieCandidate
        :return: a set of candidate strings, each contained in a TrieCandidate object
        """
        cdef list candidates
        cdef str current_name, ch2, letters, deasciified_word, replaced, added, added_last, deleted, swapped
        cdef int current_index, j
        cdef float current_penalty
        cdef TrieNode current_node
        candidates = []
        current_name = candidate.getName()
        current_index = candidate.getCurrentIndex()
        current_penalty = candidate.getCurrentPenalty()
        current_node = self.__trie.getTrieNode(current_name[:current_index])
        if current_node is None:
            return candidates
        letters = current_node.childrenToString()
        if current_name[current_index] == 'c':
            ch2 = 'ç'
        elif current_name[current_index] == 'g':
            ch2 = 'ğ'
        elif current_name[current_index] == 'i':
            ch2 = 'ı'
        elif current_name[current_index] == 's':
            ch2 = 'ş'
        elif current_name[current_index] == 'o':
            ch2 = 'ö'
        elif current_name[current_index] == 'u':
            ch2 = 'ü'
        else:
            ch2 = ''
        if ch2 != '':
            deasciified_word = current_name[:current_index] + ch2 + current_name[current_index + 1:]
        else:
            deasciified_word = current_name
        if deasciified_word != current_name and self.__trie.startsWith(deasciified_word[:current_index + 1]):
            candidates.append(TrieCandidate(deasciified_word, current_index, current_penalty + 0.2))
        for j in range(len(letters)):
            replaced = current_name[:current_index] + letters[j] + current_name[current_index + 1:]
            candidates.append(TrieCandidate(replaced, current_index, current_penalty + 1))
            added = current_name[:current_index] + letters[j] + current_name[current_index:]
            candidates.append(TrieCandidate(added, current_index, current_penalty + 1))
            added_last = current_name + letters[j]
            if self.__trie.startsWith(added_last):
                candidates.append(TrieCandidate(added_last, current_index, current_penalty + 1))
        deleted = current_name[:current_index] + current_name[current_index + 1:]
        if len(deleted) > 1:
            candidates.append(TrieCandidate(deleted, current_index - 1, current_penalty + 1))
        if current_index < len(current_name) - 1 and current_name[current_index] != current_name[current_index + 1] \
                and self.__trie.startsWith(current_name[:current_index + 2]):
            swapped = current_name[:current_index] + current_name[current_index + 1] + current_name[current_index] + \
                      current_name[current_index + 2:]
            candidates.append(TrieCandidate(swapped, current_index, current_penalty + 1))
        return candidates



import re

from Corpus.Sentence cimport Sentence
from Dictionary.Word cimport Word
from MorphologicalAnalysis.FsmMorphologicalAnalyzer cimport FsmMorphologicalAnalyzer
from MorphologicalAnalysis.FsmParseList cimport FsmParseList
from NGram.NGram cimport NGram

from SpellChecker.Candidate cimport Candidate
from SpellChecker.Operator import Operator
from SpellChecker.SimpleSpellChecker cimport SimpleSpellChecker


cdef class NGramSpellChecker(SimpleSpellChecker):

    def __init__(self,
                 fsm: FsmMorphologicalAnalyzer,
                 nGram: NGram,
                 parameter: SpellCheckerParameter):
        """
        A constructor of NGramSpellChecker class which takes a FsmMorphologicalAnalyzer and an NGram as inputs. Then,
        calls its super class SimpleSpellChecker with given FsmMorphologicalAnalyzer and assigns given NGram to the
        nGram variable.

        PARAMETERS
        ----------
        fsm : FsmMorphologicalAnalyzer
            FsmMorphologicalAnalyzer type input.
        nGram : NGram
            NGram type input.
        """
        super().__init__(fsm)
        self.__nGram = nGram
        self.__parameter = parameter

    cpdef Word checkAnalysisAndSetRootForWordAtIndex(self,
                                                     Sentence sentence,
                                                     int index):
        """
        Checks the morphological analysis of the given word in the given index. If there is no misspelling, it returns
        the longest root word of the possible analyses.
        :param sentence Sentence to be analyzed.
        :param index Index of the word
        :return If the word is misspelled, null; otherwise the longest root word of the possible analyses.
        """
        cdef str word_name, upper_case_word_name
        cdef FsmParseList fsm_parses, upper_case_fsm_parses
        if index < sentence.wordCount():
            word_name = sentence.getWord(index).getName()
            compiled_expression1 = re.compile(".*\d+.*")
            compiled_expression2 = re.compile(".*[a-zA-ZçöğüşıÇÖĞÜŞİ]+.*")
            if (compiled_expression1.fullmatch(word_name) and compiled_expression2.fullmatch(word_name) \
                and "'" not in word_name) or len(word_name) < self.__parameter.getMinWordLength():
                return sentence.getWord(index)
            fsm_parses = self.fsm.morphologicalAnalysis(sentence.getWord(index).getName())
            if fsm_parses.size() != 0:
                if self.__parameter.isRootNGram():
                    return fsm_parses.getParseWithLongestRootWord().getWord()
                else:
                    return sentence.getWord(index)
            else:
                upper_case_word_name = Word.toCapital(word_name)
                upper_case_fsm_parses = self.fsm.morphologicalAnalysis(upper_case_word_name)
                if upper_case_fsm_parses.size() != 0:
                    if self.__parameter.isRootNGram():
                        return upper_case_fsm_parses.getParseWithLongestRootWord().getWord()
                    else:
                        return sentence.getWord(index)
        return None

    cpdef Word checkAnalysisAndSetRoot(self, str word):
        """
        Checks the morphological analysis of the given word. If there is no misspelling, it returns
        the longest root word of the possible analysis.
        :param word: Word to be analyzed.
        :return: If the word is misspelled, null; otherwise the longest root word of the possible analysis.
        """
        cdef FsmParseList fsm_parses_of_word, fsm_parses_of_capitalized_word
        fsm_parses_of_word = self.fsm.morphologicalAnalysis(word)
        if fsm_parses_of_word.size() != 0:
            if self.__parameter.isRootNGram():
                return fsm_parses_of_word.getParseWithLongestRootWord().getWord()
            else:
                return Word(word)
        fsm_parses_of_capitalized_word = self.fsm.morphologicalAnalysis(word)
        if fsm_parses_of_capitalized_word.size() != 0:
            if self.__parameter.isRootNGram():
                return fsm_parses_of_capitalized_word.getParseWithLongestRootWord().getWord()
            else:
                return Word(word)
        return None

    cpdef setThreshold(self, float threshold):
        self.__threshold = threshold

    cpdef float getProbability(self,
                               str word1,
                               str word2):
        return self.__nGram.getProbability(word1, word2)

    cpdef Sentence spellCheck(self, Sentence sentence):
        """
        The spellCheck method takes a Sentence as an input and loops i times where i ranges from 0 to size of words in
        given sentence. Then, it calls morphologicalAnalysis method with each word and assigns it to the FsmParseList,
        if the size of FsmParseList is equal to the 0, it adds current word to the candidateList and assigns it to the
        candidates list.

        Later on, it loops through candidates list and calls morphologicalAnalysis method with each word and assigns it
        to the FsmParseList. Then, it gets the root from FsmParseList. For the first time, it defines a previousRoot by
        calling getProbability method with root, and for the following times it calls getProbability method with
        previousRoot and root. Then, it finds out the best probability and the corresponding candidate as best candidate
        and adds it to the result Sentence.

        If the size of FsmParseList is not equal to 0, it directly adds the current word to the result Sentence and
        finds the previousRoot directly from the FsmParseList.

        PARAMETERS
        ----------
        sentence : Sentence
            Sentence type input.

        RETURNS
        -------
        Sentence
            Sentence result.
        """
        cdef Candidate best_candidate, candidate
        cdef Sentence result
        cdef int i, repeat
        cdef Word previous_root, word, next_word, previous_word, next_next_word
        cdef Word previous_previous_word, best_root, root, next_root
        cdef FsmParseList fsm_parses
        cdef list candidates
        cdef double best_probability, previous_probability, next_probability
        previous_root = None
        result = Sentence()
        root = self.checkAnalysisAndSetRootForWordAtIndex(sentence, 0)
        next_root = self.checkAnalysisAndSetRootForWordAtIndex(sentence, 1)
        i = 0
        while i < sentence.wordCount():
            next_word = None
            previous_word = None
            next_next_word = None
            previous_previous_word = None
            word = sentence.getWord(i)
            if i > 0:
                previous_word = sentence.getWord(i - 1)
            if i > 1:
                previous_previous_word = sentence.getWord(i - 2)
            if i < sentence.wordCount() - 1:
                next_word = sentence.getWord(i + 1)
            if i < sentence.wordCount() - 2:
                next_next_word = sentence.getWord(i + 2)
            if self.forcedMisspellCheck(word, result):
                previous_root = self.checkAnalysisAndSetRootForWordAtIndex(result, result.wordCount() - 1)
                root = next_root
                next_root = self.checkAnalysisAndSetRootForWordAtIndex(sentence, i + 2)
                i = i + 1
                continue
            if self.forcedBackwardMergeCheck(word, result, previous_word) or \
                    self.forcedSuffixMergeCheck(word, result, previous_word):
                previous_root = self.checkAnalysisAndSetRootForWordAtIndex(result, result.wordCount() - 1)
                root = self.checkAnalysisAndSetRootForWordAtIndex(sentence, i + 1)
                next_root = self.checkAnalysisAndSetRootForWordAtIndex(sentence, i + 2)
                i = i + 1
                continue
            if self.forcedForwardMergeCheck(word, result, next_word) or \
                    self.forcedHyphenMergeCheck(word, result, previous_word, next_word):
                i = i + 1
                previous_root = self.checkAnalysisAndSetRootForWordAtIndex(result, result.wordCount() - 1)
                root = self.checkAnalysisAndSetRootForWordAtIndex(sentence, i + 1)
                next_root = self.checkAnalysisAndSetRootForWordAtIndex(sentence, i + 2)
                i = i + 1
                continue
            if self.forcedSplitCheck(word, result) or self.forcedShortcutCheck(word, result):
                previous_root = self.checkAnalysisAndSetRootForWordAtIndex(result, result.wordCount() - 1)
                root = next_root
                next_root = self.checkAnalysisAndSetRootForWordAtIndex(sentence, i + 2)
                i = i + 1
                continue
            if self.__parameter.isDeMiCheck():
                if self.forcedDeDaSplitCheck(word, result) or self.forcedQuestionSuffixSplitCheck(word, result):
                    previous_root = self.checkAnalysisAndSetRootForWordAtIndex(result, result.wordCount() - 1)
                    root = next_root
                    next_root = self.checkAnalysisAndSetRootForWordAtIndex(sentence, i + 2)
                    i = i + 1
                    continue
            if root is None or \
                    (len(word.getName()) <= self.__parameter.getMinWordLength() and self.fsm.morphologicalAnalysis(word.getName()).size() == 0):
                candidates = []
                if root is None:
                    candidates.extend(self.candidateList(word, sentence))
                    candidates.extend(self.splitCandidatesList(word))
                candidates.extend(self.mergedCandidatesList(previous_word, word, next_word))
                best_candidate = Candidate(word.getName(), Operator.NO_CHANGE)
                best_root = word
                best_probability = self.__parameter.getThreshold()
                for candidate in candidates:
                    if candidate.getOperator() == Operator.SPELL_CHECK or \
                            candidate.getOperator() == Operator.MISSPELLED_REPLACE or \
                            candidate.getOperator() == Operator.TRIE_BASED or \
                            candidate.getOperator() == Operator.CONTEXT_BASED:
                        root = self.checkAnalysisAndSetRoot(candidate.getName())
                    if candidate.getOperator() == Operator.BACKWARD_MERGE and previous_word is not None:
                        root = self.checkAnalysisAndSetRoot(previous_word.getName() + word.getName())
                        if previous_previous_word is not None:
                            previous_root = self.checkAnalysisAndSetRoot(previous_previous_word.getName())
                    if candidate.getOperator() == Operator.FORWARD_MERGE and next_word is not None:
                        root = self.checkAnalysisAndSetRoot(word.getName() + next_word.getName())
                        if next_next_word is not None:
                            next_root = self.checkAnalysisAndSetRoot(next_next_word.getName())
                    if previous_root is not None:
                        if candidate.getOperator() == Operator.SPLIT:
                            root = self.checkAnalysisAndSetRoot(candidate.getName().split(" ")[0])
                        previous_probability = self.getProbability(previous_root.getName(), root.getName())
                    else:
                        previous_probability = 0.0
                    if next_root is not None:
                        if candidate.getOperator() == Operator.SPLIT:
                            root = self.checkAnalysisAndSetRoot(candidate.getName().split(" ")[1])
                        next_probability = self.getProbability(root.getName(), next_root.getName())
                    else:
                        next_probability = 0.0
                    if max(previous_probability, next_probability) > best_probability or len(candidates) == 1:
                        best_candidate = candidate
                        best_root = root
                        best_probability = max(previous_probability, next_probability)
                if best_candidate.getOperator() == Operator.FORWARD_MERGE:
                    i = i + 1
                if best_candidate.getName() == Operator.BACKWARD_MERGE:
                    result.replaceWord(i - 1, Word(best_candidate.getName()))
                else:
                    if best_candidate.getName() == Operator.SPLIT:
                        self.addSplitWords(best_candidate.getName(), result)
                    else:
                        result.addWord(Word(best_candidate.getName()))
                root = best_root
            else:
                result.addWord(word)
            previous_root = root
            root = next_root
            next_root = self.checkAnalysisAndSetRootForWordAtIndex(sentence, i + 2)
            i = i + 1
        return result



cdef class SpellCheckerParameter:

    cdef float __threshold
    cdef bint __deMiCheck
    cdef bint __rootNGram
    cdef int __minWordLength

    cpdef setThreshold(self, float threshold)
    cpdef setDeMiCheck(self, bint deMiCheck)
    cpdef setRootNGram(self, bint rootNGram)
    cpdef float getThreshold(self)
    cpdef bint isDeMiCheck(self)
    cpdef bint isRootNGram(self)
    cpdef setMinWordLength(self, int minWordLength)
    cpdef int getMinWordLength(self)



import pkg_resources
from Corpus.Sentence cimport Sentence
from Dictionary.Word cimport Word
from MorphologicalAnalysis.FsmMorphologicalAnalyzer cimport FsmMorphologicalAnalyzer
from MorphologicalAnalysis.FsmParseList cimport FsmParseList
from NGram.NGram cimport NGram

from SpellChecker.Candidate cimport Candidate
from SpellChecker.NGramSpellChecker cimport NGramSpellChecker
from SpellChecker.Operator import Operator
from SpellChecker.SpellCheckerParameter cimport SpellCheckerParameter


cdef class ContextBasedSpellChecker(NGramSpellChecker):

    __context_list: dict

    def __init__(self,
                 fsm: FsmMorphologicalAnalyzer,
                 nGram: NGram,
                 parameter: SpellCheckerParameter):
        """
        A constructor of {@link ContextBasedSpellChecker} class which takes a {@link FsmMorphologicalAnalyzer}, an {@link NGram}
        and a {@link SpellCheckerParameter} as inputs. Then, calls its super class {@link NGramSpellChecker} with given inputs.
        :param fsm: {@link FsmMorphologicalAnalyzer} type input.
        :param nGram: {@link NGram} type input.
        :param parameter: {@link SpellCheckerParameter} type input.
        """
        super().__init__(fsm,
                         nGram,
                         parameter)
        self.loadContextDictionaries()

    cpdef loadContextDictionaries(self):
        """
        This method also loads context information from a file.
        """
        cdef list lines, items, otherWords
        cdef str line, word
        self.__context_list = {}
        input_file = open(pkg_resources.resource_filename(__name__, 'data/context_list.txt'), "r", encoding="utf8")
        lines = input_file.readlines()
        for line in lines:
            items = line.strip().split("\t")
            if len(items) == 2:
                word = items[0]
                otherWords = items[1].split(" ")
                self.__context_list[word] = otherWords
        input_file.close()

    cpdef list candidateList(self, Word word, Sentence sentence):
        """
        Uses context information to generate candidates for a misspelled word.
        The candidates are the words that are in the context of the neighbouring words of the misspelled word.
        Uses the {@Link damerauLevenshteinDistance(String, String) method to calculate the distance between the misspelled word and
        the candidates and to determine whether the candidates are valid.
        :param word: the misspelled word
        :param sentence: the sentence containing the misspelled word
        :return: an ArrayList of valid candidates for the misspelled word
        """
        cdef list words, valid_candidates
        cdef set candidates
        cdef Word w
        cdef FsmParseList parses
        cdef str root, s
        cdef Candidate candidate
        cdef int distance
        words = []
        candidates = set()
        valid_candidates = []
        for w in sentence.getWords():
            if w != word:
                words.append(w)
        for w in words:
            parses = self.fsm.morphologicalAnalysis(Word.toCapital(w.getName()))
            if parses.size() > 0:
                root = parses.getParseWithLongestRootWord().getWord().getName()
                if self.__context_list.get(root) is not None:
                    for s in self.__context_list[root]:
                        candidates.add(Candidate(s, Operator.CONTEXT_BASED))
        for candidate in candidates:
            if len(candidate.getName()) < 5:
                distance = 1
            elif len(candidate.getName()) < 7:
                distance = 2
            else:
                distance = 3
            if self.damerauLevenshteinDistance(word.getName(), candidate.getName()) <= distance:
                valid_candidates.append(candidate)
        return valid_candidates

    cpdef int damerauLevenshteinDistance(self, str first, str second):
        """
        Calculates the Damerau-Levenshtein distance between two strings.
        This method also allows for the transposition of adjacent characters,
        which is not possible in a standard Levenshtein distance calculation.
        :param first: the first string
        :param second: the second string
        :return: the Damerau-Levenshtein distance between the two strings
        """
        cdef int first_length, second_length, first_index, second_index, cost
        cdef list distance_matrix
        if first is None or first == "":
            if second is None or second == "":
                return 0
            else:
                return len(second)
        if second is None or second == "":
            return len(first)
        first_length = len(first)
        second_length = len(second)
        distance_matrix = []
        for first_index in range(first_length + 1):
            distance_matrix.append([])
            for second_index in range(second_length + 1):
                distance_matrix[first_index].append(0)
        for first_index in range(first_length + 1):
            distance_matrix[first_index][0] = first_index
        for second_index in range(second_length + 1):
            distance_matrix[0][second_index] = second_index
        for first_index in range(1, first_length + 1):
            for second_index in range(1, second_length + 1):
                if first[first_index - 1] == second[second_index - 1]:
                    cost = 0
                else:
                    cost = 1
                distance_matrix[first_index][second_index] = min(min(distance_matrix[first_index - 1][second_index] + 1,
                                                                  distance_matrix[first_index][second_index - 1] + 1),
                                                              distance_matrix[first_index - 1][second_index - 1] + cost)
                if first_index == 1 or second_index == 1:
                    continue
                if first[first_index - 1] == second[second_index - 2] and first[first_index - 2] == second[
                    second_index - 1]:
                    distance_matrix[first_index][second_index] = min(distance_matrix[first_index][second_index],
                                                                  distance_matrix[first_index - 2][
                                                                      second_index - 2] + cost)
        return distance_matrix[first_length][second_length]
